"""
### Generic utility functions ###

Functions that make interacting with the toolbox easier.
"""

# External
import numpy as np
from numba import jit


def channel_first_to_last(arr: np.ndarray) -> np.ndarray:
    """
    Converts a numpy array from channel first to channel last format.
    `(-batch-, channel, height, width)` -> `(-batch-, height, width, channel)`
    
    If 4D, it is assumed that the input array is in batch, channel, height, width format.
    If 3D, it is assumed that the input array is in channel, height, width format.

    Parameters
    ----------
    arr : np.ndarray
        The array to convert.

    Returns
    -------
    np.ndarray
        The converted array.
    """
    if arr.ndim not in [3, 4]:
        raise ValueError("Input array should be 3 or 4-dimensional with shape (-batch-, channels, height, width)")

    # Swap the axes to change from channel first to channel last format
    if arr.ndim == 3:
        arr = np.transpose(arr, (1, 2, 0))
    else:
        arr = np.transpose(arr, (0, 2, 3, 1))

    return arr


def channel_last_to_first(arr: np.ndarray) -> np.ndarray:
    """
    Converts a numpy array from channel last to channel first format.

    `(-batch-, height, width, channel)` -> `(-batch-, channel, height, width)`
    
    If 4D, it is assumed that the input array is in batch, channel, height, width format.
    If 3D, it is assumed that the input array is in channel, height, width format.

    Parameters
    ----------
    arr : np.ndarray
        The array to convert.

    Returns
    -------
    np.ndarray
        The converted array.
    """
    if arr.ndim not in [3, 4]:
        raise ValueError("Input array should be 3 or 4-dimensional with shape (-batch-, height, width, channels)")

    # Swap the axes to change from channel last to channel first format
    if arr.ndim == 3:
        arr = np.transpose(arr, (2, 0, 1))
    else:
        arr = np.transpose(arr, (0, 3, 1, 2))

    return arr


@jit(nopython=True)
def _create_grid(range_rows, range_cols):
    """
    Create a grid of rows and columns.
    
    Parameters
    ----------
    range_rows : np.ndarray
        The rows to create the grid from.

    range_cols : np.ndarray
        The columns to create the grid from.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The rows and columns grid.
    """
    rows_grid = np.zeros((len(range_rows), len(range_cols)), dtype=np.int64)
    cols_grid = np.zeros((len(range_rows), len(range_cols)), dtype=np.int64)

    for i in range(len(range_rows)):
        for j in range(len(range_cols)):
            cols_grid[i, j] = range_rows[j]
            rows_grid[i, j] = range_cols[i]

    return rows_grid, cols_grid



"""
This module contains functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
import sys; sys.path.append("../../")
from typing import Optional, List, Callable, Union

# External
import numpy as np

# Buteo
from buteo.array.utils_array import channel_first_to_last, channel_last_to_first
from buteo.array.loaders import MultiArray
from buteo.ai.augmentation_funcs import (
    augmentation_mirror,
    augmentation_mirror_xy,
    augmentation_rotation,
    augmentation_rotation_xy,
    augmentation_noise_uniform,
    augmentation_noise_normal,
    augmentation_channel_scale,
    augmentation_contrast,
    augmentation_drop_channel,
    augmentation_drop_pixel,
    augmentation_blur,
    augmentation_blur_xy,
    augmentation_sharpen,
    augmentation_sharpen_xy,
    augmentation_cutmix,
    augmentation_mixup,
    augmentation_misalign
)


class AugmentationDataset():
    """
    A dataset that applies augmentations to the data.
    Every augmentation added needs to be specified as a dictionary
    with the following keys:
        - name (str)
        - chance (float[0,1])

    The following augmentations are supported:
        - rotation
        - mirror
        - channel_scale (additive(bool), max_value(float[0,1]))
        - noise (additive(bool), max_value(float[0,1]))
        - contrast (max_value(float[0,1]))
        - drop_pixel (drop_probability(float[0,1]), drop_value(float))
        - drop_channel (drop_probability(float[0,1]), drop_value(float))
        - blur
        - blur_xy
        - sharpen
        - sharpen_xy
        - cutmix (min_size(float[0, 1]), max_size(float[0, 1]))
        - mixup (min_size(float[0, 1]), max_size(float[0, 1]), label_mix(int))

    Parameters
    ----------
    X : np.ndarray
        The data to augment.

    y : np.ndarray
        The labels for the data.

    augmentations : list, optional
        The augmentations to apply.

    callback : callable, optional
        A callback to apply to the data after augmentation.

    input_is_channel_last : bool, default: True
        Whether the data is in channel last format.

    output_is_channel_last : bool, default: False
        Whether the output should be in channel last format.

    Returns
    -------
    AugmentationDataset
        A dataset yielding batches of augmented data. For Pytorch,
        convert the batches to tensors before ingestion.

    Example
    -------
    >>> def callback(x, y):
    ...     return (
    ...         torch.from_numpy(x).float(),
    ...         torch.from_numpy(y).float(),
    ...     )
    ...
    >>> dataset = AugmentationDataset(
    ...     x_train,
    ...     y_train,
    ...     callback=callback,
    ...     input_is_channel_last=True,
    ...     output_is_channel_last=False,
    ...     augmentations=[
    ...         { "name": "rotation", "chance": 0.2},
    ...         { "name": "mirror", "chance": 0.2 },
    ...         { "name": "noise", "chance": 0.2 },
    ...         { "name": "cutmix", "chance": 0.2 },
    ...     ],
    ... )
    >>>
    >>> from torch.utils.data import DataLoader
    >>> dataloader = DataLoader(dataset, batch_size=32, shuffle=True)
    """
    def __init__(self,
        X: Union[Union[np.ndarray, MultiArray], List[Union[np.ndarray, MultiArray]]],
        y: Union[Union[np.ndarray, MultiArray], List[Union[np.ndarray, MultiArray]]],
        augmentations: Optional[List] = None,
        callback: Callable = None,
        callback_pre: Callable = None,
        input_is_channel_last: Optional[bool] = None,
        output_is_channel_last: Optional[bool] = None,
    ):
        self.x_train = X
        self.y_train = y

        self.augmentations = augmentations or []
        self.callback = callback
        self.callback_pre = callback_pre
        self.input_is_channel_last = input_is_channel_last
        self.output_is_channel_last = output_is_channel_last

        # Read the first sample to determine if it is multi input
        self.x_is_multi_input = isinstance(self.x_train[0], list) and len(self.x_train[0]) > 1
        self.y_is_multi_input = isinstance(self.y_train[0], list) and len(self.y_train[0]) > 1

        assert len(self.x_train) == len(self.y_train), "X and y must have the same length."
        assert input_is_channel_last is not None, "Input channel format must be specified."
        assert output_is_channel_last is not None, "Output channel format must be specified."

        # If X is more than one array, then we need to make sure that the
        # number of list of augmentations is the same as the number of arrays.
        if self.x_is_multi_input or self.y_is_multi_input:
            if len(self.augmentations) == 0:
                pass
            elif not isinstance(self.augmentations[0], list):
                x_len = len(self.x_train[0]) if self.x_is_multi_input else 1
                y_len = len(self.y_train[0]) if self.y_is_multi_input else 1

                self.augmentations = [self.augmentations] * (x_len + y_len)
            else:
                assert len(self.augmentations) == len(X), "Number of augmentations must match number of arrays."

        test_augs = [self.augmentations] if not (self.x_is_multi_input or self.y_is_multi_input) else self.augmentations
        for aug in test_augs:
            if not self._validate_augmentations(aug):
                raise ValueError("Augmentations are invalid.")

    def __len__(self):
        return len(self.x_train)
    
    def _validate_augmentations(self, augmentations) -> bool:
        """ Validates the augmentations. """
        multi_input = self.x_is_multi_input or self.y_is_multi_input

        if len(augmentations) == 0:
            return True
        
        for aug in augmentations:
                # Check if augmentation is valid
            if "name" not in aug:
                return False

            if "p" not in aug and "chance" not in aug:
                return False

            aug_name = aug["name"]
            if "chance" in aug:
                aug_change = aug["chance"]
            elif "p" in aug:
                aug_change = aug["p"]

            if aug_change is None:
                return False
            
            if not 0 <= aug_change <= 1:
                return False
            
            if aug_name is None:
                return False

            # Check if augmentation is valid for multi input
            if (aug_name[-2:] == "xy" or aug_name in ["cutmix", "mixup"]) and multi_input:
                return False
            
        return True

    def _apply_augmentations(self,
        sample_x: List[np.ndarray],
    ) -> List[np.ndarray]:
        """ Applies the augmentations to the sample. Inplace. Only x is possible. """
        # Apply augmentations
        for i in range(len(sample_x)):
            for aug in self.augmentations[i]:
                aug_name = aug["name"]
                if "chance" in aug:
                    aug_change = aug["chance"]
                elif "p" in aug:
                    aug_change = aug["p"]
                func = None

                # Check if augmentation should be applied
                if np.random.rand() > aug_change:
                    break

                # Mapping augmentation names to their respective functions
                if aug_name == "rotation":
                    func = augmentation_rotation
                elif aug_name == "mirror":
                    func = augmentation_mirror
                elif aug_name == "channel_scale":
                    func = augmentation_channel_scale
                elif aug_name == "noise_uniform":
                    func = augmentation_noise_uniform
                elif aug_name == "noise_normal":
                    func = augmentation_noise_normal
                elif aug_name == "contrast":
                    func = augmentation_contrast
                elif aug_name == "drop_pixel":
                    func = augmentation_drop_pixel
                elif aug_name == "drop_channel":
                    func = augmentation_drop_channel
                elif aug_name == "blur":
                    func = augmentation_blur
                elif aug_name == "sharpen":
                    func = augmentation_sharpen
                elif aug_name == "misalign":
                    func = augmentation_misalign

                if func is None:
                    raise ValueError(f"Augmentation {aug['name']} not supported.")

                kwargs = {key: value for key, value in aug.items() if key not in ["name", "chance", "inplace", "p"]}

                sample_x[i] = func(sample_x[i], channel_last=False, inplace=True, **kwargs)

        return sample_x

    def _apply_augmentations_xy(self,
        sample_x: List[np.ndarray],
        sample_y: List[np.ndarray],
    ) -> List[np.ndarray]:
        """ Applies the augmentations to the sample. Inplace. Both x and y are possible. """
        x = sample_x[0]
        y = sample_y[0]

        # Apply augmentations
        for aug in self.augmentations:
            aug_name = aug["name"]
            if "chance" in aug:
                aug_change = aug["chance"]
            elif "p" in aug:
                aug_change = aug["p"]
            func = None

            # Check if augmentation should be applied
            if np.random.rand() > aug_change:
                break

            # Mapping augmentation names to their respective functions
            if aug_name == "rotation":
                func = augmentation_rotation
            elif aug_name == "rotation_xy":
                func = augmentation_rotation_xy
            elif aug_name == "mirror":
                func = augmentation_mirror
            elif aug_name == "mirror_xy":
                func = augmentation_mirror_xy
            elif aug_name == "channel_scale":
                func = augmentation_channel_scale
            elif aug_name == "noise_uniform":
                func = augmentation_noise_uniform
            elif aug_name == "noise_normal":
                func = augmentation_noise_normal
            elif aug_name == "contrast":
                func = augmentation_contrast
            elif aug_name == "drop_pixel":
                func = augmentation_drop_pixel
            elif aug_name == "drop_channel":
                func = augmentation_drop_channel
            elif aug_name == "blur":
                func = augmentation_blur
            elif aug_name == "blur_xy":
                func = augmentation_blur_xy
            elif aug_name == "sharpen":
                func = augmentation_sharpen
            elif aug_name == "sharpen_xy":
                func = augmentation_sharpen_xy
            elif aug_name == "misalign":
                func = augmentation_misalign
            elif aug_name == "cutmix":
                func = augmentation_cutmix
            elif aug_name == "mixup":
                func = augmentation_mixup

            if func is None:
                raise ValueError(f"Augmentation {aug['name']} not supported.")

            kwargs = {key: value for key, value in aug.items() if key not in ["name", "chance", "inplace", "p"]}

            # Augmentations that apply to both image and label
            if aug_name in ["rotation_xy", "mirror_xy", "blur_xy", "sharpen_xy"]:
                x, y = func(x, y, channel_last=False, inplace=True, **kwargs)

            # Augmentations that needs two images
            elif aug_name in ["cutmix", "mixup"]:
                idx_source = np.random.randint(len(self.x_train))

                xx, yy = (self.x_train[idx_source], self.y_train[idx_source])

                if isinstance(xx, list): xx = xx[0]
                if isinstance(yy, list): yy = yy[0]

                # Convert the format of the input data if necessary
                if self.input_is_channel_last:
                    xx = channel_last_to_first(xx)
                    yy = channel_last_to_first(yy)

                # If using cutmix/mixup, we have to draw samples from other images.
                # This means that we have to apply the callback_pre those samples
                # before applying the augmentation.
                if self.callback_pre is not None:
                    xx, yy = self.callback_pre(xx, yy)
                
                xx = np.array(xx) if isinstance(xx, np.memmap) else xx
                yy = np.array(yy) if isinstance(yy, np.memmap) else yy

                x, y = func(x, y, xx, yy, channel_last=False, inplace=True, **kwargs)

            # Augmentations that only apply to image
            else:
                x = func(x, channel_last=False, inplace=True, **kwargs)
            
            sample_x = [x]
            sample_y = [y]
        
        return sample_x, sample_y

    def __getitem__(self, index):
        # Ensure the samples are in list format
        sample_x, sample_y = self._ensure_list_format(self.x_train[index], self.y_train[index])

        # Convert the format of the input data if necessary
        if self.input_is_channel_last:
            sample_x = [channel_last_to_first(x) for x in sample_x]
            sample_y = [channel_last_to_first(y) for y in sample_y]

        # Apply callback_pre if specified. For normalization.
        if self.callback_pre is not None:
            sample_x, sample_y = self._apply_callback(self.callback_pre, sample_x, sample_y)

        sample_x = [np.array(x) if isinstance(x, np.memmap) else x.copy() for x in sample_x]
        sample_y = [np.array(y) if isinstance(y, np.memmap) else y.copy() for y in sample_y]

        # Apply augmentations
        if self.x_is_multi_input or self.y_is_multi_input:
            sample_x = self._apply_augmentations(sample_x)
        else:
            sample_x, sample_y = self._apply_augmentations_xy(sample_x, sample_y)

        # Apply callback if specified
        if self.callback is not None:
            sample_x, sample_y = self._apply_callback(self.callback, sample_x, sample_y)

        # Convert the format of the output data if necessary
        if self.output_is_channel_last:
            sample_x = [channel_first_to_last(x) for x in sample_x]
            sample_y = [channel_first_to_last(y) for y in sample_y]

        # Restore the samples to their original format
        return self._restore_original_format(sample_x), self._restore_original_format(sample_y)

    def _ensure_list_format(self, x, y):
        return (x if isinstance(x, list) else [x]), (y if isinstance(y, list) else [y])

    def _copy_and_convert_format(self, data, format_flag, converter_func):
        return converter_func(data) if format_flag else data

    def _apply_callback(self, callback_func, sample_x, sample_y):
        preconv_x, preconv_y = sample_x[0], sample_y[0]
        callback_x, callback_y = callback_func(preconv_x, preconv_y)
        return [callback_x], [callback_y]

    def _restore_original_format(self, data):
        return data[0] if len(data) == 1 else data


class Dataset:
    """
    A dataset that does not apply any augmentations to the data.
    Allows a callback to be passed and can convert between
    channel formats.

    Parameters
    ----------
    X : np.ndarray or MultiArray
        The data to read.

    y : np.ndarray or MultiArray
        The labels for the data.

    callback : callable, optional
        A callback to apply to the data before returning.
        Inside the callback, the format will always be channel first.

    input_is_channel_last : bool, default: True
        Whether the data is in channel last format.

    output_is_channel_last : bool, default: False
        Whether the output should be in channel last format.

    Returns
    -------
    Dataset
        A dataset to iterate over combining X and y.
    """
    def __init__(self,
        X: Union[Union[np.ndarray, MultiArray], List[Union[np.ndarray, MultiArray]]],
        y: Union[Union[np.ndarray, MultiArray], List[Union[np.ndarray, MultiArray]]],
        callback: Callable = None,
        input_is_channel_last: Optional[bool] = True,
        output_is_channel_last: Optional[bool] = False,
    ):
        assert len(X) == len(y), "X and y must have the same length."
        self.x_train = X
        self.y_train = y
        self.callback = callback
        self.input_is_channel_last = input_is_channel_last
        self.output_is_channel_last = output_is_channel_last

    def __len__(self):
        return len(self.x_train)

    def __getitem__(self, index):
        sample_x, sample_y = self._ensure_list_format(self.x_train[index], self.y_train[index])

        # Copy and optionally convert the format of the input data
        sample_x = [self._copy_and_convert_format(x, self.input_is_channel_last, channel_last_to_first) for x in sample_x]
        sample_y = [self._copy_and_convert_format(y, self.input_is_channel_last, channel_last_to_first) for y in sample_y]

        sample_x = [np.array(x) if isinstance(x, np.memmap) else x.copy() for x in sample_x]
        sample_y = [np.array(y) if isinstance(y, np.memmap) else y.copy() for y in sample_x]

        # Apply callback if specified
        if self.callback is not None:
            sample_x, sample_y = self._apply_callback(sample_x, sample_y)

        # Convert the output format if necessary
        sample_x = [self._copy_and_convert_format(x, self.output_is_channel_last, channel_first_to_last) for x in sample_x]
        sample_y = [self._copy_and_convert_format(y, self.output_is_channel_last, channel_first_to_last) for y in sample_y]

        return self._restore_original_format(sample_x), self._restore_original_format(sample_y)

    def _ensure_list_format(self, x, y):
        return (x if isinstance(x, list) else [x]), (y if isinstance(y, list) else [y])

    def _copy_and_convert_format(self, data, format_flag, converter_func):
        return converter_func(data) if format_flag else data

    def _apply_callback(self, sample_x, sample_y):
        preconv_x, preconv_y = sample_x[0], sample_y[0]
        callback_x, callback_y = self.callback(preconv_x, preconv_y)
        return [callback_x], [callback_y]

    def _restore_original_format(self, data):
        return data[0] if len(data) == 1 else data



from typing import Union, List, Tuple

import numpy as np


class MultiArray:
    def __init__(self,
        array_list: Union[np.ndarray, List[Union[np.ndarray, List[np.ndarray]]]]
    ):
        """
        This is a class that takes in a tuple of list of arrays and glues them together
        without concatenating them. This is useful for when you have a large
        dataset that you want to load into memory, but you don't want to
        concatenate them because that would take up too much memory.

        The function works for saved numpy arrays loading using mmap_mode="r".

        Parameters
        ----------
        array_list : Union[np.ndarray, List[Union[np.ndarray, List[np.ndarray]]]]
            The input can be multimodal data in the following formats:
                1. A single numpy array. 
                2. A list of numpy arrays. (single modality) - will be iterable
                3. A list of lists of numpy arrays. (multi-modality) - Each list will be iterable, returning a tuple of arrays.

        Returns
        -------
        MultiArray
            A class that can handle multiple memmapped numpy arrays.

        Examples
        --------
        ```python
        >>> from glob import glob
        >>> folder = "./data_patches/"

        >>> patches = sorted(glob(folder + "train*.npy"))
        >>> multi_array = MultiArray([np.load(p, mmap_mode="r") for p in patches])
        >>> single_image = multi_array[0]

        >>> print(single_image.shape)
        (128, 128, 10)
        >>> print(len(multi_array))
        32
        ```
        """
        self.list_list_arrays, self.original_format = self._convert_input(array_list)
        self.cumulative_lens = []

        # Calculate the cumulative lengths once for all the arrays
        for i in range(len(self.list_list_arrays)):
            self.cumulative_lens.append(
                np.cumsum([0] + [arr.shape[0] for arr in self.list_list_arrays[i]])
            )

        # Check that all arrays have the same length (across modalities)
        if len(self.cumulative_lens) > 1:
            for i in range(1, len(self.cumulative_lens)):
                if not np.array_equal(self.cumulative_lens[i], self.cumulative_lens[i-1]):
                    raise ValueError("All arrays should have the same length.")

        # They should all be the same, so just take the first one
        self.cumulative_sizes = self.cumulative_lens[0]


    def _convert_input(self,
        array_list: Union[np.ndarray, List[Union[np.ndarray, List[np.ndarray]]]],
    ) -> Tuple[List[List[np.ndarray]], str]:
        """
        The input can be:
            1. A single numpy array.
            2. A list of numpy arrays 
            3. A list of lists of numpy arrays

        All inputs will be converted to format 3. The output will be in the original format.
        """
        # Case 1: single numpy array
        if isinstance(array_list, np.ndarray):
            return [[array_list]], 'array'  

        elif isinstance(array_list, list):
            if not array_list:
                raise ValueError("Input list is empty. Please provide a list with numpy arrays.")

            # Case 2: list of numpy arrays
            if all(isinstance(item, np.ndarray) for item in array_list):
                return [array_list], 'list'

            # Case 3: list of lists of numpy arrays
            elif all(isinstance(item, list) and all(isinstance(sub_item, np.ndarray) for sub_item in item) for item in array_list):
                return array_list, 'list_list'

            else:
                raise ValueError("Input list should either contain numpy arrays or lists of numpy arrays.")
        else:
            raise TypeError("Invalid input type. Input should either be a numpy array or a list of numpy arrays or a list of lists of numpy arrays.")


    def _revert_input(self,
        converted_array_list: List[List[np.ndarray]], original_format: str,
    ) -> Union[np.ndarray, List[Union[np.ndarray, List[np.ndarray]]]]:
        """ Convert the list of lists back to the original format. """
        if original_format == 'array':
            return converted_array_list[0][0]

        elif original_format == 'list':
            return converted_array_list[0]

        elif original_format == 'list_list':
            return converted_array_list
        else:
            raise ValueError("Invalid format string. It should be one of the following: 'array', 'list', 'list_list'.")


    def __getitem__(self, idx: int):
        if idx < 0:  # support negative indexing
            idx = self.__len__() + idx
        elif idx >= self.__len__():
            raise IndexError("Index out of range")

        array_idx = np.searchsorted(self.cumulative_sizes, idx, side='right') - 1

        output = [
            self.list_list_arrays[i][array_idx][idx - self.cumulative_sizes[array_idx]]
            for i in range(len(self.list_list_arrays))
        ]

        return self._revert_input(output, self.original_format)


    def __len__(self):
        return self.cumulative_sizes[-1]



"""
This module contains functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
import sys; sys.path.append("../../")
import random
from typing import Optional, Tuple, Any

# External
import numpy as np
from numba import jit, prange

# Internal
from buteo.ai.augmentation_utils import (
    _rotate_arr,
    _mirror_arr,
)
from buteo.array.convolution import (
    convolve_array_simple,
)
from buteo.array.convolution_kernels import (
    _simple_blur_kernel_2d_3x3,
    _simple_unsharp_kernel_2d_3x3,
    _simple_shift_kernel_2d,
)


def augmentation_rotation(
    X: np.ndarray,
    k: int = -1,
    channel_last: bool = True,
    inplace = False,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:

    if not inplace:
        X = X.copy()

    if k == -1:
        random_k = random.choice([1, 2, 3])
    else:
        random_k = k

    X[:] = _rotate_arr(X, random_k, channel_last)

    return X


def augmentation_rotation_xy(
    X: np.ndarray,
    y: np.ndarray,
    k: int = -1,
    channel_last: bool = True,
    inplace: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Randomly rotate the image and label by 90 degrees intervals. Images
    can be (channels, height, width) or (height, width, channels).
    
    Parameters
    ----------
    X : np.ndarray
        The image to rotate.

    y : np.ndarray
        The label to rotate.

    k : int, optional
        The number of 90 degree intervals to rotate by, default: -1 (random).

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The rotated image and optionally the label.
    """
    if not inplace:
        X = X.copy()
        y = y.copy()

    if k == -1:
        random_k = random.choice([1, 2, 3])
    else:
        random_k = k

    X[:] = _rotate_arr(X, random_k, channel_last)
    y[:] = _rotate_arr(y, random_k, channel_last)

    return X, y


def augmentation_mirror(
    X: np.ndarray,
    k: int = -1,
    channel_last: bool = True,
    inplace: bool = False,
) -> np.ndarray:
    """
    Randomly mirrors the image.
    Images can be (channels, height, width) or (height, width, channels).
    
    Parameters
    ----------
    X : np.ndarray
        The image to mirror.

    k : int, optional
        If -1, randomly mirrors the image along the horizontal or vertical axis.
        1. mirrors the image along the horizontal axis.
        2. mirrors the image along the vertical axis.
        3. mirrors the image along both the horizontal and vertical axis, default: None.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The mirrored image.
    """
    if not inplace:
        X = X.copy()

    if k == -1:
        random_k = random.choice([1, 2, 3])
    else:
        random_k = k

    X[:] = _mirror_arr(X, random_k, channel_last)

    return X


def augmentation_mirror_xy(
    X: np.ndarray,
    y: np.ndarray,
    k: int = -1,
    channel_last: bool = True,
    inplace: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Randomly mirrors the image. Images can be (channels, height, width) or (height, width, channels).
    
    Parameters
    ----------
    X : np.ndarray
        The image to mirror.

    y : np.ndarray
        The label to mirror.

    k : int, optional
        If -1, randomly mirrors the image along the horizontal or vertical axis.
        1. mirrors the image along the horizontal axis.
        2. mirrors the image along the vertical axis.
        3. mirrors the image along both the horizontal and vertical axis, default: None.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The mirrored image and optionally the label.
    """
    if not inplace:
        X = X.copy()
        y = y.copy()

    if k == -1:
        random_k = random.choice([1, 2, 3])
    else:
        random_k = k

    X[:] = _mirror_arr(X, random_k, channel_last)
    y[:] = _mirror_arr(y, random_k, channel_last)

    return X, y


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_noise_uniform(
    X: np.ndarray,
    max_amount: float = 0.1,
    additive: bool = False,
    channel_last: Any = None, # pylint: disable=unused-argument
    inplace: bool = False,
) -> np.ndarray:
    """
    Adds random noise seperately to each pixel of the image. The noise works
    for both channel first and last images. Follows a uniform distribution.
    Input should be (height, width, channels) or (channels, height, width).

    NOTE: Beware of datatypes. Consider casting to float32 before adding noise.
    
    Parameters
    ----------
    X : np.ndarray
        The image to add noise to.

    max_amount : float, optional
        The maximum amount of noise to add, sampled uniformly, default: 0.1.

    additive : bool, optional
        Whether to add or multiply the noise, default: False.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), ignored for this function.
        Kept to keep the same function signature as other augmentations, default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The image with uniform noise.
    """
    if not inplace:
        X = X.copy()

    amount = np.random.rand() * max_amount

    if additive:
        random_noise = np.random.uniform(-amount, amount, X.shape).astype(X.dtype)
        X[:] += random_noise
    else:
        random_noise = np.random.uniform(1 - amount, 1 + amount, X.shape).astype(X.dtype)
        X[:] *= random_noise

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_noise_normal(
    X: np.ndarray,
    max_amount: float = 0.1,
    additive: bool = False,
    channel_last: Any = None, # pylint: disable=unused-argument
    inplace: bool = False,
) -> np.ndarray:
    """
    Adds random noise seperately to each pixel of the image. The noise works
    for both channel first and last images. Follows a normal distribution.
    max_amount is the standard deviation of the normal distribution.
    Input should be (height, width, channels) or (channels, height, width).

    NOTE: Beware of datatypes. Consider casting to float32 before adding noise.
    
    Parameters
    ----------
    X : np.ndarray
        The image to add noise to.

    max_amount : float, optional
        The maximum amount of noise to add, sampled uniformly, default: 0.1.

    additive : bool, optional
        Whether to add or multiply the noise, default: False.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), ignored for this function.
        Kept to keep the same function signature as other augmentations, default: True.

    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The image with normal noise.
    """
    if not inplace:
        X = X.copy()

    amount = np.random.rand() * max_amount

    if additive:
        random_noise = np.random.normal(0, amount, X.shape).astype(X.dtype)
        X[:] += random_noise
    else:
        random_noise = np.random.normal(1, amount, X.shape).astype(X.dtype)
        X[:] *= random_noise

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_channel_scale(
    X: np.ndarray,
    max_amount: float = 0.1,
    additive: bool = False,
    channel_last: bool = True,
    inplace: bool = False,
) -> np.ndarray:
    """
    Scales the channels of the image seperately by a fixed amount.
    Input should be (height, width, channels) or (channels, height, width).

    NOTE: Beware of datatypes. Consider casting to float32 before adding noise.
    
    Parameters
    ----------
    X : np.ndarray
        The image to scale the channels of.

    max_amount : float, optional
        The amount to possible scale the channels by. Sampled uniformly, default: 0.1.

    additive : bool, optional
        Whether to add or multiply the scaling, default: False.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The image with scaled channels.
    """
    if not inplace:
        X = X.copy()

    amount = np.random.rand() * max_amount

    if channel_last:
        for i in prange(X.shape[2]):
            if additive:
                random_amount = np.random.uniform(-amount, amount)
                X[:, :, i] += random_amount
            else:
                random_amount = np.random.uniform(1 - amount, 1 + amount)
                X[:, :, i] *= random_amount
    else:
        for i in prange(X.shape[0]):
            if additive:
                random_amount = np.random.uniform(-amount, amount)
                X[i, :, :] += random_amount
            else:
                random_amount = np.random.uniform(1 - amount, 1 + amount)
                X[i, :, :] *= random_amount

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_contrast(
    X: np.ndarray,
    max_amount: float = 0.1,
    channel_last: bool = True,
    inplace: bool = False,
) -> np.ndarray:
    """
    Changes the contrast of an image by a random amount, seperately for each channel.
    Input should be (height, width, channels) or (channels, height, width).
    
    NOTE: Beware of datatypes. Consider casting to float32 before adding noise.

    Parameters
    ----------
    X : np.ndarray
        The image to change the contrast of.

    max_amount : float, optional
        The max amount to change the contrast by, default: 0.1.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The image with changed contrast.
    """
    if not inplace:
        X = X.copy()

    amount = np.random.rand() * max_amount

    channels = X.shape[2] if channel_last else X.shape[0]
    mean_pixel = np.zeros(channels, dtype=np.float32)

    # Numba workaround because numba does not support axis argument in np.mean
    if channel_last:
        for i in prange(channels):
            mean_pixel[i] = np.nanmean(X[:, :, i])
    else:
        for i in prange(channels):
            mean_pixel[i] = np.nanmean(X[i, :, :])

    for i in prange(channels):
        X[:, :, i] = (X[:, :, i] - mean_pixel[i]) * (1 + amount) + mean_pixel[i]

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_drop_pixel(
    X: np.ndarray,
    drop_probability: float = 0.01,
    drop_value: float = 0.0,
    channel_last: bool = True,
    inplace: bool = False,
) -> np.ndarray:
    """
    Drops random pixels from an image. Input should be (height, width, channels) or (channels, height, width).
    Only drops pixels from features, not labels.
    
    Parameters
    ----------
    X : np.ndarray
        The image to drop a pixel from.

    drop_probability : float, optional
        The probability of dropping a pixel, default: 0.05.

    drop_value : float, optional
        The value to drop the pixel to, default: 0.0.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The image with dropped pixels.
    """
    if not inplace:
        X = X.copy()

    # Create a mask of random numbers
    mask = np.random.uniform(0.0, 1.0, size=X.shape).astype(np.float32)

    # Agreed. This looks terrible. But it's the only way to get numba to parallelize this.
    if channel_last:
        for col in prange(X.shape[0]):
            for row in prange(X.shape[1]):
                for channel in prange(X.shape[2]):
                    if mask[col, row, channel] <= drop_probability:
                        X[col, row, channel] = drop_value

    else:
        for channel in prange(X.shape[0]):
            for col in prange(X.shape[1]):
                for row in prange(X.shape[2]):
                    if mask[channel, col, row] <= drop_probability:
                        X[channel, col, row] = drop_value

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_drop_channel(
    X: np.ndarray,
    drop_value: float = 0.0,
    channel_last: bool = True,
    inplace: bool = False,
) -> np.ndarray:
    """
    Drops a random channel from an image. Input should be (height, width, channels) or (channels, height, width).
    A maximum of one channel will be dropped.
    
    Parameters
    ----------
    X : np.ndarray
        The image to drop a channel from.

    drop_value : float, optional
        The value to drop the channel to, default: 0.0.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The image with dropped channels.
    """
    if not inplace:
        X = X.copy()

    channels = X.shape[2] if channel_last else X.shape[0]
    channel_to_drop = np.random.randint(0, channels)

    if channel_last:
        X[:, :, channel_to_drop] = drop_value
    else:
        X[channel_to_drop, :, :] = drop_value

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_blur(
    X: np.ndarray,
    channel_last: bool = True,
    inplace: bool = False,
) -> np.ndarray:
    """
    Blurs an image at random. Input should be (height, width, channels) or (channels, height, width).
    
    NOTE: Beware of datatypes. Consider casting to float32 before adding noise.

    Parameters
    ----------
    X : np.ndarray
        The image to blur.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The blurred image.
    """
    if not inplace:
        X = X.copy()

    offsets, weights = _simple_blur_kernel_2d_3x3()

    if channel_last:
        for channel in prange(X.shape[2]):
            X[:, :, channel] = convolve_array_simple(
                X[:, :, channel],
                offsets,
                weights
            )
    else:
        for channel in prange(X.shape[0]):
            X[channel, :, :] = convolve_array_simple(
                X[channel, :, :],
                offsets,
                weights,
            )

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_blur_xy(
    X: np.ndarray,
    y: np.ndarray,
    channel_last: bool = True,
    inplace: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Blurs an image at random. Input should be (height, width, channels) or (channels, height, width).
    The label is blurred by the same amount.

    NOTE: Beware of datatypes. Consider casting to float32 before adding noise.
    
    Parameters
    ----------
    X : np.ndarray
        The image to blur.

    y : np.ndarray
        The label to blur.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The blurred image and optionally the unmodified label.
    """
    x_blurred, _ = augmentation_blur(X, channel_last, inplace)
    y_blurred, _ = augmentation_blur(y, channel_last, inplace)

    return x_blurred, y_blurred


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_sharpen(
    X: np.ndarray,
    channel_last: bool = True,
    inplace: bool = False,
) -> np.ndarray:
    """
    Sharpens an image at random. Input should be (height, width, channels) or (channels, height, width).

    NOTE: Beware of datatypes. Consider casting to float32 before adding noise.

    Parameters
    ----------
    X : np.ndarray
        The image to sharpen.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.

    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    np.ndarray
        The sharpened image.
    """
    if not inplace:
        X = X.copy()

    offsets, weights = _simple_unsharp_kernel_2d_3x3()

    if channel_last:
        for channel in prange(X.shape[2]):
            X[:, :, channel] = convolve_array_simple(
                X[:, :, channel],
                offsets,
                weights
            )
    else:
        for channel in prange(X.shape[0]):
            X[channel, :, :] = convolve_array_simple(
                X[channel, :, :],
                offsets,
                weights,
            )

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_sharpen_xy(
    X: np.ndarray,
    y: np.ndarray,
    channel_last: bool = True,
    inplace: bool = False,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Sharpens an image at random. Input should be (height, width, channels) or (channels, height, width).
    The label is sharpened by the same amount.

    NOTE: Beware of datatypes. Consider casting to float32 before adding noise.

    Parameters
    ----------
    X : np.ndarray
        The image to sharpen.

    y : np.ndarray
        The label to sharpen. If None, no label is returned

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The sharpened image and label.
    """
    x_sharpened, _ = augmentation_sharpen(X, channel_last, inplace)
    y_sharpened, _ = augmentation_sharpen(y, channel_last, inplace)

    return x_sharpened, y_sharpened


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_misalign(
    X: np.ndarray,
    max_offset: float = 0.5,
    channel_last: bool = True,
    inplace: bool = False,
) -> np.ndarray:
    """
    Misaligns one channel in the image at random.
    input should be (height, width, channels) or (channels, height, width).

    Parameters
    ----------
    X : np.ndarray
        The image to misalign the channels of.

    max_offset : float, optional
        The maximum offset to misalign the channels by. Default: 0.5.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels). Default: True.

    Returns
    -------
    np.ndarray
        The misaligned image.
    """
    if not inplace:
        X = X.copy()

    offsets, weights = _simple_shift_kernel_2d(
        min(np.random.rand(), max_offset),
        min(np.random.rand(), max_offset),
    )

    channels = X.shape[2] if channel_last else X.shape[0]
    channel_to_adjust = np.random.randint(0, channels)

    if channel_last:
        X[:, :, channel_to_adjust] = convolve_array_simple(
            X[:, :, channel_to_adjust], offsets, weights,
        )
    else:
        X[channel_to_adjust, :, :] = convolve_array_simple(
            X[channel_to_adjust, :, :], offsets, weights,
        )

    return X


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_cutmix(
    X_target: np.ndarray,
    y_target: np.ndarray,
    X_source: np.ndarray,
    y_source: np.ndarray,
    min_size: float = 0.333,
    max_size: float = 0.666,
    channel_last: bool = True,
    inplace: bool = False,
) -> Tuple[np.ndarray, np.ndarray,]:
    """
    Cutmixes two images.
    Input should be (height, width, channels) or (channels, height, width).

    Parameters
    ----------
    X_target : np.ndarray
        The image to transfer the cutmix to.

    y_target : np.ndarray
        The label to transfer the cutmix to.

    X_source : np.ndarray
        The image to cutmix from.

    y_source : np.ndarray
        The label to cutmix from.

    min_size : float, optional
        The minimum size of the patch to cutmix. In percentage of the image width, default: 0.333.

    max_size : float, optional
        The maximum size of the patch to cutmix. In percentage of the image width, default: 0.666.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The cutmixed image and label.
    """
    if not inplace:
        X_target = X_target.copy()
        y_target = y_target.copy()

    if channel_last:
        height, width, _channels_x = X_target.shape
    else:
        _channels_x, height, width = X_target.shape

    # Create random size patch
    patch_height = np.random.randint(int(height * min_size), int(height * max_size))
    patch_width = np.random.randint(int(width * min_size), int(width * max_size))

    # Determine patch location
    x0 = np.random.randint(0, width - patch_width)
    y0 = np.random.randint(0, height - patch_height)
    x1 = x0 + patch_width
    y1 = y0 + patch_height

    # Cut and paste
    if channel_last:
        X_target[y0:y1, x0:x1, :] = X_source[y0:y1, x0:x1, :]
        y_target[y0:y1, x0:x1, :] = y_source[y0:y1, x0:x1, :]

    else:
        X_target[:, y0:y1, x0:x1] = X_source[:, y0:y1, x0:x1]
        y_target[:, y0:y1, x0:x1] = y_source[:, y0:y1, x0:x1]

    return X_target, y_target


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_mixup(
    X_target: np.ndarray,
    y_target: np.ndarray,
    X_source: np.ndarray,
    y_source: np.ndarray,
    min_size: float = 0.333,
    max_size: float = 0.666,
    label_mix: int = 0,
    channel_last: bool = True, # pylint: disable=unused-argument
    inplace: bool = False,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Mixups two images at random. This works by doing a linear intepolation between
    two images and then adding a random weight to each image.

    Mixup involves taking two images and blending them together by randomly interpolating
    their pixel values. More specifically, suppose we have two images x and x' with their
    corresponding labels y and y'. To generate a new training example, mixup takes a
    weighted sum of x and x', such that the resulting image x^* = λx + (1-λ)x',
    where λ is a randomly chosen interpolation coefficient. The label for the new image
    is also a weighted sum of y and y' based on the same interpolation coefficient.

    input should be (height, width, channels) or (channels, height, width).

    Parameters
    ----------
    X_target : np.ndarray
        The image to transfer to.

    y_target : np.ndarray
        The label to transfer to.

    X_source : np.ndarray
        The image to transfer from.

    y_source : np.ndarray
        The label to transfer from.

    min_size : float, optional
        The minimum mixup coefficient, default: 0.333.

    max_size : float, optional
        The maximum mixup coefficient, default: 0.666.

    label_mix : int, optional
        If 0, the labels will be mixed by the weights. If 1, the target label will be used. If 2, 
        the source label will be used. If 3, the max of the labels will be used. If 4, the min 
        of the labels will be used. If 5, the max of the image with the highest weight will be used. 
        If 6, the min of the image with the highest weight will be used. If 7, the sum of the labels 
        will be used, default: 0.

    channel_last : bool, optional
        Whether the image is (channels, height, width) or (height, width, channels), default: True.
    
    inplace : bool, optional
        Whether to perform the rotation in-place, default: False.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The mixed up image and label.
    """
    if not inplace:
        X_target = X_target.copy()
        y_target = y_target.copy()

    mixup_coeff = np.float32(
        min(np.random.uniform(min_size, max_size + 0.001), 1.0),
    )

    X_target[:] = (X_target * mixup_coeff) + (X_source * (np.float32(1.0) - mixup_coeff))

    if label_mix == 0:
        y_target[:] = (y_target * mixup_coeff) + (y_source * (np.float32(1.0) - mixup_coeff))
    elif label_mix == 1:
        y_target[:] = y_target # pylint: disable=self-assigning-variable
    elif label_mix == 2:
        y_target[:] = y_source
    elif label_mix == 3:
        y_target[:] = np.maximum(y_target, y_source)
    elif label_mix == 4:
        y_target[:] = np.minimum(y_target, y_source)
    elif label_mix == 5:
        y_target[:] = y_target if mixup_coeff >= 0.5 else y_source
    elif label_mix == 6:
        y_target[:] = y_target if mixup_coeff >= 0.5 else y_source
    elif label_mix == 7:
        y_target[:] = y_target + y_source

    return X_target, y_target
